package polyglot.ext.predj;

import polyglot.lex.Lexer;
import polyglot.ext.predj.parse.Lexer_c;
import polyglot.ext.predj.parse.Grm;
import polyglot.ext.predj.ast.PredJNodeFactory_c;
import polyglot.ext.predj.types.PredJTypeSystem_c;
import polyglot.ext.predj.visit.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.frontend.*;
import polyglot.main.*;

import java.util.*;
import java.io.*;

/**
 * Extension information for predj extension.
 */
public class ExtensionInfo extends polyglot.ext.jl5.ExtensionInfo {
    static {
        // force Topics to load
        Topics t = new Topics();
    }

    public String defaultFileExtension() {
        return "pj";
    }

    public String compilerName() {
        return "predjc";
    }

    public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        Lexer lexer = new Lexer_c(reader, source.name(), eq);
        Grm grm = new Grm(lexer, ts, nf, eq);
        return new CupParser(grm, source, eq);
    }

    protected NodeFactory createNodeFactory() {
        return new PredJNodeFactory_c();
    }

    protected TypeSystem createTypeSystem() {
        return new PredJTypeSystem_c();
    }

    public static final Pass.ID PREDICATE_POLARITY =
    new Pass.ID("predicate-polarity");

    public static final Pass.ID PURITY_CHECK =
    new Pass.ID("check-purity");

    public static final Pass.ID CANONICALIZE_PREDICATES =
    new Pass.ID("canonicalize-predicates");

    public static final Pass.ID CREATE_PREDICATE_TYPES =
    new Pass.ID("create-predicate-types");

    public static final Pass.ID CREATE_PREDICATE_TYPES_ALL =
    new Pass.ID("create-predicate-types-barrier");
    
    public static final Pass.ID EXTRACT_BINDINGS =
    new Pass.ID("extract-bindings");

    public static final Pass.ID BUILD_DISPATCHERS =
    new Pass.ID("build-dispatchers");

    public static final Pass.ID IMPLEMENTATION_SIDE_TYPE_CHECK =
    new Pass.ID("implementation-side-type-check");
    
    public static final Pass.ID TRANSLATE_PREDJ =
    new Pass.ID("translate-predj");


	
//JJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
//JPRED TEAM CODE--------------------------------------------------------------------J
//JJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ	

    // Added for Predj Eclipse AST data extraction
	public static final Pass.ID PREDJ_ECLIPSE_PASS =
		new Pass.ID("predj-eclipse-pass");

//-----------------------------------------------------------------------------------	
	
    public List passes(Job job) {
        List passes = super.passes(job);

	    // compute predicate polarity immediately (right after parsing)
	    // this determines whether bindings from this predicate should
	    // be allowed to escape for use in the body of the method
	afterPass(passes, Pass.PARSE,
		   new VisitorPass(PREDICATE_POLARITY,
				   job, new PredicatePolarity()));

            // make sure that pure methods are actually pure, and that
            // only pure methods are called inside predicates
        afterPass(passes, polyglot.ext.jl5.ExtensionInfo.TYPE_CHECK_ALL,
                   new VisitorPass(PURITY_CHECK, job,
                                   new PurityChecker(job, ts, nf)));

	    // canonicalize each method predicate.  this relies on
	    // info computed in the typecheck pass
	afterPass(passes, PURITY_CHECK,
		   new VisitorPass(CANONICALIZE_PREDICATES, job,
				   new PredicateCanonicalizer(job, ts, nf)));

	afterPass(passes, CANONICALIZE_PREDICATES,
		   new VisitorPass(CREATE_PREDICATE_TYPES, job,
				   new PredicateTypeCreator(job, ts, nf)));

	
	    // this barrier is necessary so that when we read the source
	    // file of some depended-upon class whose file is not on the
	    // command line, we will process the file up to the
	    // CREATE_PREDICATE_TYPES pass, which sets the predicate
	    // type information for each method instance in the class.
	    // otherwise, ITC on the depending class can crash.
	afterPass(passes, CREATE_PREDICATE_TYPES,
		  new BarrierPass(CREATE_PREDICATE_TYPES_ALL, job));
	
	    /* The Build_Dispatchers pass unfortunately cannot come before
	       typechecking, because it needs field instances to have
	       already been set, which is done in the typeCheck pass.
	       Field instances are used to qualify field names with their
	       containers, so we can distinguish fields of the same name
	       in the subclass and superclass.  This is critical for
	       getting a correct partial order among methods (and for ITC,
	       performed in a later pass).

	       The Extract_Bindings pass also needs type information.
	    */

	    // extract the extra and refined bindings used in the body
	    // of a predicate method; these are needed for properly
	    // translating the method to Java
	afterPass(passes, CREATE_PREDICATE_TYPES_ALL,
		  new VisitorPass(EXTRACT_BINDINGS, job,
				  new BindingsExtractor()));
	
	    // create repositories for each set of dispatcher mates,
	    // and compute the partial order among dispatcher mates.
	    // each repository will eventually be turned into a dispatcher
	    // method
	afterPass(passes, EXTRACT_BINDINGS,
		   new VisitorPass(BUILD_DISPATCHERS, job,
				   new DispatcherBuilder(job, ts, nf)));

	    // perform completeness and ambiguity checking on each set
	    // of dispatcher mates.
	afterPass(passes, BUILD_DISPATCHERS,
		  new VisitorPass(IMPLEMENTATION_SIDE_TYPE_CHECK, job,
				  new ImplementationSideTypeChecker(
				      job, ts, nf)));



	    // now translate PredJ AST to Java AST
	beforePass(passes, Pass.PRE_OUTPUT_ALL,
		   new VisitorPass(TRANSLATE_PREDJ, job,
				   new PredJTranslator(job, ts, nf)));


	
//	JJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ
//	JPRED TEAM CODE--------------------------------------------------------------------J
//	JJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJJ	
	
	// Added for Eclipse AST data extraction
	// Pass will run before the Translate_Predj pass
	beforePass(passes, TRANSLATE_PREDJ,
		   new VisitorPass(PREDJ_ECLIPSE_PASS, job,
				   new PredJEclipseVisitor(job, ts, nf)));

        return passes;
    }

}
