package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.predj.types.Predicate;
import polyglot.visit.*;
import polyglot.util.*;

// a predicate that binds a local name to a field or formal
public class BindingNode_c extends BindingOrSpecializerNode_c
    implements BindingNode {

	// the stuff for local bindings is unfortunately duplicated
	// in BindingSpecializerNode_c, because we lack multiple
	// implementation inheritance
    protected PredicateVarBinding localBinding;
    
    public BindingNode_c(Position pos,
			 PredicateVarBinding localBinding,
			 Expr target) {
	super(pos, target);
	this.localBinding = localBinding;
    }
   
    public PredicateVarBinding localBinding() {
	return localBinding;
    }

    public BindingNode localBinding(PredicateVarBinding localBinding) {
	BindingNode_c n = (BindingNode_c) copy();
	n.localBinding = localBinding;
	return n;
    }

    public boolean equals(Object o) {
	if (o instanceof BindingNode) {
	    BindingNode bn = (BindingNode) o;
	    return localBinding().name().equals(bn.localBinding().name()) &&
		target().equals(bn.target());
	}
	return super.equals(o);
    }

    public Node visitChildren(NodeVisitor v) {
	PredicateVarBinding localBinding =
	    (PredicateVarBinding) visitChild(this.localBinding, v);
	Expr target = (Expr) visitChild(this.target, v);
	return reconstruct(localBinding, target);
    }

    protected BindingNode_c reconstruct(
	PredicateVarBinding localBinding, Expr target) {
	if (localBinding != this.localBinding || target != this.target) {
	    BindingNode_c n = (BindingNode_c) copy();
	    n.localBinding = localBinding;
	    n.target = target;
	    return n;
	}

	return this;
    }

    public BindingOrSpecializerNode polarity(boolean polarity) {
	BindingNode n = (BindingNode) super.polarity(polarity);

	    // pass the polarity information on to the localBinding
	return n.localBinding(n.localBinding().polarity(polarity));
    }

	/* update the context to reflect the new var binding
	 */
    public void addDecls(Context c)  {
	if (c instanceof PredJContext) {
	    PredJContext pc = (PredJContext) c;
	    if (this.polarity) {
		pc.addPredicateVarBinding(localBinding.name(), target, false);
	    }
	} else
	    super.addDecls(c);
    }
}
