package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;

import java.util.*;

// a base class for binding and specializer predicate nodes
public abstract class BindingOrSpecializerNode_c extends PredJExpr_c
    implements BindingOrSpecializerNode {

    protected Expr target;
    
	// true if the specialized type should be used in typechecking
	// the body of the associated method
    protected boolean polarity;
				
    public BindingOrSpecializerNode_c(Position pos, Expr target) {
	super(pos);
	this.target = target;
    }

    public Expr target() {
	return target;
    }

    public boolean polarity() {
	return polarity;
    }
    
    public BindingOrSpecializerNode polarity(boolean polarity) {
	BindingOrSpecializerNode_c n = (BindingOrSpecializerNode_c) copy();
	n.polarity = polarity;
	return n;
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
	BindingOrSpecializerNode n =
	    (BindingOrSpecializerNode) super.typeCheck(tc);

	    // specializers and bindings have type boolean
	return n.type(tc.typeSystem().Boolean());
    }

    
    /**
     * Return the first (sub)term performed when evaluating this
     * term.
     */   
    public Term entry() {
	return target;
    }
	
    /**
     * Visit this term in evaluation order.
     */
    public List acceptCFG(CFGBuilder v, List succs) {
	v.visitCFG(target, this);
	return succs;
    }
}
