package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;

// a specializer that binds a local name to the specializee
public class BindingSpecializerNode_c extends AbstractSpecializerNode_c
    implements BindingSpecializerNode {

    protected PredicateVarBinding localBinding;
				
    public BindingSpecializerNode_c(Position pos,
				    PredicateVarBinding localBinding,
				    Expr target,
				    TypeNode specializer) {
	super(pos, target, specializer);
	this.localBinding = localBinding;
    }

    public PredicateVarBinding localBinding() {
	return localBinding;
    }

    public BindingNode localBinding(
	PredicateVarBinding localBinding) {
	BindingSpecializerNode_c n = (BindingSpecializerNode_c) copy();
	n.localBinding = localBinding;
	return n;
    }

    public boolean equals(Object o) {
	if (o instanceof BindingSpecializerNode) {
	    BindingSpecializerNode sn = (BindingSpecializerNode) o;
	    return localBinding().name().equals(sn.localBinding().name()) &&
		target().equals(sn.target()) &&
		specializer().equals(sn.specializer());
	}
	return super.equals(o);
    }

    public Node visitChildren(NodeVisitor v) {
	PredicateVarBinding localBinding =
	    (PredicateVarBinding) visitChild(this.localBinding, v);
	Expr target = (Expr) visitChild(this.target, v);
	TypeNode specializer = (TypeNode) visitChild(this.specializer, v);
	return reconstruct(localBinding, target, specializer);
    }

    protected BindingSpecializerNode_c reconstruct(
	PredicateVarBinding localBinding,
	Expr target, TypeNode specializer) {
	if (localBinding != this.localBinding || target != this.target ||
	    specializer != this.specializer) {
	    BindingSpecializerNode_c n = (BindingSpecializerNode_c) copy();
	    n.localBinding = localBinding;
	    n.target = target;
	    n.specializer = specializer;
	    return n;
	}

	return this;
    }

    public BindingOrSpecializerNode polarity(boolean polarity) {
	BindingNode n = (BindingNode) super.polarity(polarity);

	    // pass the polarity information on to the localBinding
	return n.localBinding(n.localBinding().polarity(polarity));
    }

	/* update the context to reflect the new var binding
	 */
    public void addDecls(Context c)  {
	if (c instanceof PredJContext) {
	    PredJContext pc = (PredJContext) c;
	    if (this.polarity) {
		pc.addPredicateVarBinding(localBinding.name(), target, true);
	    }
	} else
	    super.addDecls(c);
    }

}
