package polyglot.ext.predj.ast;

import java.util.*;

import polyglot.util.*;
import polyglot.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.visit.*;

import polyglot.ext.jl5.ast.*;

import polyglot.ext.jl5.types.FlagAnnotations;

// a version of ClassDecl that supports ITC

public class PredJClassDecl_c extends JL5ClassDecl_c implements PredJClassDecl {

    public PredJClassDecl_c(Position pos, FlagAnnotations flags, String name,
			    TypeNode superClass, List interfaces,
			    ClassBody body) {
	super(pos, flags, name, superClass, interfaces, body);
    }

    public PredJClassDecl_c(Position pos, FlagAnnotations flags, String name,
			    TypeNode superClass, List interfaces,
			    ClassBody body, List paramTypes) {
	super(pos, flags, name, superClass, interfaces, body, paramTypes);
    }

	// check that all inherited gfs are completely implemented, if this
	// class is concrete.  abstract inherited methods are already checked
	// by the regular typechecks on concrete classes in Java.  this check
	// handles the case when an abstract class partially implements a gf
	// with concrete predicate methods, but does not fully implement it.
	// further, we only need deal with those gfs for which the current
	// class does not provide any overriding methods, because otherwise
	// the other checks on this class (e.g. the ITC on dispatcher mates)
	// ensures completeness.
    public Node implementationSideTypeCheck(
	ImplementationSideTypeChecker itc) throws SemanticException {

	PredJTypeSystem ts = (PredJTypeSystem) itc.typeSystem();
	
	ParsedClassType ct = (ParsedClassType) this.type();

	    // only check concrete classes
	if (ct.flags().isAbstract() || ct.flags().isInterface()) {
	    return this;
	}
	
        // build up a list of abstract superclasses that ct extends, where
	// there is no intervening concrete superclass
        List superClasses = new LinkedList();
	Type currCls = ct.superType();
	while (currCls != null && currCls.toClass().flags().isAbstract()) {
	    superClasses.add(currCls);
	    currCls = currCls.toClass().superType();
	}

        PredJParsedClassType[] absSups = (PredJParsedClassType[])
	    superClasses.toArray(new PredJParsedClassType[0]);

	for (int i = 0; i < absSups.length; i++) {
		// get the dispatchers of this class
	    List dispatchers = absSups[i].dispatchers();
	    Iterator iter = dispatchers.iterator();
	    while (iter.hasNext()) {
		PredJDispatcherInstance di =
		    (PredJDispatcherInstance) iter.next();
		    // we need to check completeness of this dispatcher
		    // if there are no overriding methods of the same gf
		    // in the current class or in intervening superclasses.
		if (!ts.isOverridden(di, ct)) {
		    boolean isOverridden = false;
		    for (int j = 0; j < i; j++) {
			if (ts.isOverridden(di, absSups[j])) {
			    isOverridden = true;
			    break;
			}
		    }

			// if the dispatcher has an abstract method,
			// then we know the concrete subclass is incomplete,
			// since we required the abstract class to be
			// incomplete for this gf.  if the dispatcher
			// does not have an abstract method, then we
			// required the dispatcher to be complete, so
			// everything is fine.
		    if (!isOverridden && di.hasAbstractMethod()) {
			throw new SemanticException(
			    "This class does not fully implement the " +
			    "operation that includes the method at " +
			    ((PredJMethodInstance) di.dispatcherMates().get(0))
			    .position().nameAndLineString() + ".",
			    ct.position());
		    }
		}
	    }
	}   
	
	return this;
    }
}
