package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.util.*;
import java.util.*;
import polyglot.ext.jl5.ast.*;
import polyglot.ext.jl5.types.FlagAnnotations;

/**
 * NodeFactory for predj extension.
 */
public class PredJNodeFactory_c extends JL5NodeFactory_c
    implements PredJNodeFactory {

    public PredJClassDecl JL5ClassDecl(Position pos, FlagAnnotations flags,
				       String name, TypeNode superClass,
				       List interfaces, ClassBody body,
	                               List paramTypes) {
	if (paramTypes == null)
	    return new PredJClassDecl_c(pos, flags, name, superClass,
					interfaces, body);
	else
	    return new PredJClassDecl_c(pos, flags, name, superClass,
					interfaces, body, paramTypes);
    }

    public JL5ClassBody JL5ClassBody(Position pos, List members){
        return new PredJClassBody_c(pos, members);
    }
    
    public PredJFormal JL5Formal(Position pos, FlagAnnotations flags,
				 TypeNode type, String name) {
        return new PredJFormal_c(pos, flags, type, name);
    }

    public PredJFormal JL5Formal(Position pos, FlagAnnotations flags,
				 TypeNode type, String name, boolean variable) {
        return new PredJFormal_c(pos, flags, type, name, variable);
    }

    public PredJMethodDecl JL5MethodDecl(Position pos, FlagAnnotations flags,
					 TypeNode returnType, String name,
					 List formals, List throwTypes,
					 Block body, List typeParams) {
	return new PredJMethodDecl_c(pos, flags, returnType, name, formals,
				     throwTypes, body, typeParams);
    }

    public PredJPredMethodDecl PredJPredMethodDecl(
	Position pos, FlagAnnotations flags, TypeNode returnType,
	String name, List argTypes, List throwConstraints,
	Block body, List typeParams, PredicateNode predicate) {
	return new PredJPredMethodDecl_c(pos, flags, returnType, name,
					 argTypes, throwConstraints, body,
					 typeParams, predicate);
    }

    public PredJBinary PredJBinary(Position pos, Expr left,
					   Binary.Operator op, Expr right) {
	return new PredJBinary_c(pos, left, op, right);
    }
    
    public PredicateNode PredicateNode(Position pos, Expr predicateExpr) {
	return new PredicateNode_c(pos, predicateExpr);
    }

    public PredJDispatcher PredJDispatcher(List dispatcherMates,
					   PredJTypeSystem ts) {
	return new PredJDispatcher_c(dispatcherMates, ts);
    }

    public SpecializerNode SpecializerNode(
	Position pos, Expr target,
	TypeNode specializer) {
	return new SpecializerNode_c(pos, target, specializer);
    }

    public BindingSpecializerNode BindingSpecializerNode(
	Position pos, Expr target,
	TypeNode specializer, PredicateVarBinding localBinding) {
	return new BindingSpecializerNode_c(pos, localBinding, target,
					    specializer);
    }
 
    public BindingNode BindingNode(
	Position pos, Expr target,
	PredicateVarBinding localBinding) {
	return new BindingNode_c(pos, localBinding, target);
    }
 
    public PredicateVarBinding PredicateVarBinding(Position pos,
						   Flags flags,
						   TypeNode type,
						   String name,
						   Expr init) {
	return new PredicateVarBinding_c(pos, flags, type, name, init);
    }
        
}
