package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.visit.*;
import polyglot.util.*;

// a variable binding in a predicate
public class PredicateVarBinding_c extends LocalDecl_c
    implements PredicateVarBinding {
	// true if the binding should be used in typechecking
	// the body of the associated method
    protected boolean polarity;

	// the type may be null, because there is no syntactic type
	// given for a regular var binding in a predicate.
    public PredicateVarBinding_c(Position pos, Flags flags, TypeNode type,
				 String name, Expr init) {
	super(pos, flags, type, name, init);
    }

    public boolean polarity() {
	return polarity;
    }

    public PredicateVarBinding polarity(boolean polarity) {
	PredicateVarBinding_c n = (PredicateVarBinding_c) copy();
	n.polarity = polarity;
	return n;
    }

    public Context enterScope(Node child, Context c) {
	PredJContext cNew = (PredJContext) super.enterScope(child, c);
	    // hack: LocalDecl_c.java adds this variable to the context
	    // when entering the scope of the var's initializer.
	    // we promptly remove it from the context.
	LocalInstance li = localInstance();
	String name = li.name();
	VarInstance vi = cNew.findVariableInThisScope(name);
	if (!polarity() && vi != null && vi.equals(li)) {
	    cNew.removeVariableFromThisScope(localInstance());
	}
	return cNew;
    }
    
    public void addDecls(Context c) {
	    // only add the variable to the context if the polarity is true
	if (polarity()) {
	    c.addVariable(localInstance());
	}
    }

    public Node disambiguate(AmbiguityRemover ar) throws SemanticException {
	    // hack: only call super, which creates the local instance,
	    // if the type is non-null.  if it's null, the local instance
	    // will be created in the typeCheck pass below, at which time
	    // the initializer's type will be available for use.
	    // it's null for binding nodes, non-null for binding specializer
	    // nodes (because the specializer is the type).
	if (type() != null) {
	    return super.disambiguate(ar);
	}
	return this;
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
	if (type() == null) {
	    PredicateVarBinding n = this;
		// set the type node
	    n = (PredicateVarBinding) n.type(
		tc.nodeFactory().CanonicalTypeNode(null, n.init().type()));
		// set the local instance
	    n = (PredicateVarBinding) n.localInstance(
		tc.typeSystem().localInstance(n.position(), n.flags(),
					      n.declType(), n.name()));
	    return n.del().typeCheck(tc);
	} else {
	    return super.typeCheck(tc);
	}
    }
}
