package polyglot.ext.predj.ast;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.jl5.types.*;

// a specializer for a formal parameter
public class SpecializerNode_c extends AbstractSpecializerNode_c
    implements SpecializerNode {
				
    public SpecializerNode_c(Position pos, Expr target,
			     TypeNode specializer) {
	super(pos, target, specializer);
    }

    public boolean equals(Object o) {
	if (o instanceof SpecializerNode_c) {
	    SpecializerNode sn = (SpecializerNode) o;
	    return target().equals(sn.target()) &&
		specializer().equals(sn.specializer());
	} else
	    return super.equals(o);
    }

    public int hashCode() {
	return target().hashCode() + specializer.hashCode();
    }

    public Node visitChildren(NodeVisitor v) {
	Expr target = (Expr) visitChild(this.target, v);
	TypeNode specializer = (TypeNode) visitChild(this.specializer, v);
	return reconstruct(target, specializer);
    }

    protected SpecializerNode_c reconstruct(
	Expr target, TypeNode specializer) {
	if (target != this.target || specializer != this.specializer) {
	    SpecializerNode_c n = (SpecializerNode_c) copy();
	    n.target = target;
	    n.specializer = specializer;
	    return n;
	}

	return this;
    }


    public boolean specializeType() {
	    // allow the specialized type to be used when typechecking
	    // the body if the target is a formal (i.e. not a field)
	Expr t = target();
	return t instanceof Local;
    }
    
  public Node typeCheck(TypeChecker tc) throws SemanticException {
      Expr target = target();
      
      if (target instanceof Local) {
	  PredJContext c = (PredJContext) tc.context();
	  String formalName = ((Local) target).name();
	  if (c.formalIsSpecialized(formalName) && this.polarity) {
		  // we've already refined this formal's type!
	      throw new SemanticException(
		  "The formal parameter " + formalName +
		  " is specialized more than once.", position());
	  }
      }
	
      return super.typeCheck(tc);
  }
    
	/* update the context to reflect the specialized type of the target
	 */
    public void addDecls(Context c) {
	if (c instanceof PredJContext) {
	    PredJContext pc = (PredJContext) c;
	    // update the context with the more refined type info,
	    // if this is a formal parameter specializer and the
	    // polarity is positive.
	    if (this.polarity && this.specializeType()) {
		LocalInstance li = ((Local) target).localInstance();
		pc.addRefinedFormalType(li, specializer.type());
	    }
	} else
	    super.addDecls(c);
    }

	// what position is the specialized formal parameter (if this
	// specializee is a formal) in the list of formals?  should
	// only be called after typechecking
    public int paramPosition() {
	PredJFormalInstance fi =
	    (PredJFormalInstance) ((Local) target).localInstance();
	return fi.paramPosition();
    }
}
