package polyglot.ext.predj.types;

import polyglot.types.*;
import polyglot.ext.jl.types.*;
import polyglot.util.*;
import java.util.*;

import polyglot.ext.jl5.types.*;

/* A base class for the PredJ method instance classes. */

public class PredJAbstractMethodInstance_c extends JL5MethodInstance_c
    implements PredJMethodInstance {

 	// for deserializing types
    protected PredJAbstractMethodInstance_c() {
    }
    
    public PredJAbstractMethodInstance_c(TypeSystem ts, Position pos,
					 ReferenceType container,
					 Flags flags, Type returnType,
					 String name,
					 List formalTypes, List excTypes) {
        super(ts, pos, container, flags, returnType, name,
	      formalTypes, excTypes);
    }

	// does this method have a predicate?
    public boolean hasPredicate() {
	return false;
    }

     
	/** By inheriting implementedImpl, we are considering a method
	    with a predicate to override all other methods of that gf
	    in this class or superclasses, for the purposes of
	    checking various well-formedness properties between a
	    method and those it overrides (see canOverride()).
	    Because of our encapsulated style, a method does override
	    all methods of the gf in a superclass.  As for dispatcher
	    mates (those in the same class), this semantics allows us
	    to enforce that they all have the same flags and
	    exceptions. **/
	        
	/** There are also methods overrides and overridesImpl in
	    MethodInstance_c.java.  They do the same thing as implemented
	    and implementedImpl, respectively, but they don't search
	    interfaces.  They don't seem to be used at all, so I've not
	    overridden them here. **/    

    
/** We require that all dispatcher mates have the same flags and
    exceptions.  It is possible to relax this as in Java, but it's not
    clear how clients can easily make use of this relaxation.  We'd
    have to statically figure out when only some predicate method or an
    overridden method can be invoked at some call site.
    We do make one slight relaxation and allow the existence of an
    abstract method as a mate.
**/
    public boolean canOverrideImpl(MethodInstance mi, boolean quiet)
	throws SemanticException {
	PredJMethodInstance pmi = (PredJMethodInstance) mi;
	if (hasDispatcherMate(pmi)) {
	    if (! throwsSameSet(pmi)) {
		if (quiet) return false;
		throw new SemanticException("Dispatcher mates " + signature() +
					    " and " + pmi.signature() +
					    " must have the same throwset",
					    position());
	    }

	    if (! flags().clearAbstract().equals(
		pmi.flags().clearAbstract())) {
		if (quiet) return false;
		throw new SemanticException("Dispatcher mates " + signature() +
					    " and " + pmi.signature() +
					    " must have identical flags",
					    position());
	    }
	}

	return super.canOverrideImpl(mi, quiet);
    }

    public boolean equalsImpl(TypeObject m) {
	return m == this;
    }

    public boolean throwsSameSet(PredJMethodInstance mi) {
	return this.throwsSubset(mi) && mi.throwsSubset(this);
    }

	// a dispatcher mate is a method with the same name, formal argument
	// types, and receiver.
    public boolean hasDispatcherMate(PredJMethodInstance mj) {
	return name().equals(mj.name()) && this.hasFormals(mj.formalTypes())
	    && container().equals(mj.container());
    }
}
