
package polyglot.ext.predj.types;

import polyglot.types.*;
import polyglot.ext.jl.types.Context_c;

import polyglot.ext.jl5.types.JL5Context_c;

import polyglot.ast.Expr;

import java.util.*;

/** A context that can locally refine the types of variables and fields in
    scope, based on specializer information. **/

public class PredJContext_c extends JL5Context_c implements PredJContext {

	// keeps track of the refined types (Types) of formal parameters
	// (VarInstances)
    protected Map refinedFormalTypes = new HashMap();

	// keeps track of the definitions (Exprs) of var bindings in
	// predicates (Strings)
    protected Map predicateVarBindings = new HashMap();
	// keeps track of those var bindings whose definitions are
	// specdialized
    protected Set refinedVars = new HashSet();

    public PredJContext_c(TypeSystem ts) {
	super(ts);
    }

	/** Adds a mapping from a formal to its refined type. **/
    public void addRefinedFormalType(VarInstance var, Type t) {
	refinedFormalTypes.put(var, t);
    }

	// does this formal parameter have a specialized type in the
	// current context?
    public boolean formalIsSpecialized(String name) {
	Iterator iter = refinedFormalTypes.keySet().iterator();
	while (iter.hasNext()) {
	    VarInstance vi = (VarInstance) iter.next();
	    if (name.equals(vi.name())) {
		return true;
	    }
	}
	return false;
    }
    
    protected VarInstance refineFormalType(VarInstance var) {
	Type t = (Type) refinedFormalTypes.get(var);	
	if (t == null) {
	    return var;
	} else {
	    VarInstance v = (VarInstance) var.copy();
	    v.setType(t);
	    return v;
	}
    }

    public void addPredicateVarBinding(String name, Expr def,
				       boolean isSpecialized) {
	predicateVarBindings.put(name, def);
	if (isSpecialized) {
	    refinedVars.add(name);
	}
    }

    public Expr getPredicateVarBinding(String name) {
	return (Expr) predicateVarBindings.get(name);
    }

    public boolean varBindingIsSpecialized(String name) {
	return refinedVars.contains(name);
    }
    
    public Context pushCode(CodeInstance ci) {
	PredJContext_c c = (PredJContext_c) super.pushCode(ci);
	    // when entering the scope of a new method or constructor,
	    // remove the refined type bindings of the previous one
	c.refinedFormalTypes = new HashMap();
	return c;
    }
	
    public VarInstance findVariableSilent(String name) {
	VarInstance var = super.findVariableSilent(name);
	if (var == null)
	    return null;
	return refineFormalType(var);
   }

	// remove a variable binding
    public void removeVariableFromThisScope(VarInstance var) {
	if (vars == null) {
	    return;
	}
	vars.remove(var.name());
    }
}
