package polyglot.ext.predj.types;

import polyglot.types.*;
import polyglot.ast.Binary;
import polyglot.ast.Unary;
import polyglot.ast.Special;
import polyglot.util.*;
import java.util.*;
import polyglot.ext.jl5.types.JL5TypeSystem;

public interface PredJTypeSystem extends JL5TypeSystem {

	// this is used to make the protected version of this method
	// in TypeSystem_c publicly available
    public boolean isAccessible(MemberInstance mi, ClassType ct);

    public boolean isOverridden(PredJDispatcherInstance di,
				ClassType ct);
    
    public PredJFormalInstance formalInstance(Position pos, Flags flags,
					      Type type, String name,
					      int paramPos);
    
    public PredJPredMethodInstance predMethodInstance(Position pos,
						      ReferenceType container,
						      Flags flags,
						      Type returnType,
						      String name,
						      List argTypes,
						      List excTypes,
						      Predicate predicate);

    public PredJDispatcherInstance dispatcherInstance(
	List mates, boolean hasAbstractMethod);

    public Specializer createSpecializer(Predicate target,
					 Type specializer,
					 boolean polarity);
    
    public UnknownPredicate unknownPredicate();

    public TruePredicate truePredicate();

    public PredicateTarget createPredicateTarget(String name, Type type);
    public PredicateCall createPredicateCall(String name, List arguments,
                                             Type returnType,
                                             List argumentTypes);
    public PredicateLit createPredicateLit(Object value, Type type);
    public PredicateBinary createPredicateBinary(
	Predicate left, Binary.Operator op, Predicate right, Type type);
    public PredicateArrayAccess createPredicateArrayAccess(
	Predicate array, Predicate index, Type type);
    public PredicateUnary createPredicateUnary(
	Unary.Operator op, Predicate base, Type type);
    public PredicateSpecial createPredicateSpecial(
	Special.Kind kind, Type type);
    
    	// is the method's predicate satisfiable?
    public boolean satisfiablePredicate(PredJMethodInstance m);
    
    	// does one dispatcher mate override another?
    public boolean overrides(PredJMethodInstance mi,
			     PredJMethodInstance mj);

	// get a list of the concrete methods that are part of the given
	// method's gf and are accessible from the given class.  these
	// are the set of methods that define this class's view of the gf,
	// which is used for both typechecking and codegen of the class.
    public List accessibleConcreteMethods(ClassType ct,
					  MethodInstance mi);
    
	// is the GF consisting of the given methods (a list of
	// MethodInstances) complete?
    public boolean isCompletelyImplemented(List mis);

	// are the two methods disjoint, i.e. never both applicable at once
    public boolean areDisjoint(PredJMethodInstance mi,
			       PredJMethodInstance mj);

	// do the set of methods that override the two given ones
	// completely cover the intersection of the given ones?  we
	// assume that neither given mate overrides the other.
    public boolean haveResolvingMethods(int mate1, int mate2,
				       List[] partialOrder,
				       PredJMethodInstance[] mates);
				       
    
    public TheoremProverInterface prover();

    public Predicate predicate(PredJMethodInstance mi);

    public Flags Pure();
}
