package polyglot.ext.predj.types;

import polyglot.frontend.Source;
import polyglot.types.*;
import polyglot.ext.jl.types.TypeSystem_c;
import polyglot.ext.jl5.types.JL5TypeSystem_c;
import polyglot.util.*;
import polyglot.ast.Binary;
import polyglot.ast.Unary;
import polyglot.ast.Special;
import java.util.*;

public class PredJTypeSystem_c extends JL5TypeSystem_c
    implements PredJTypeSystem {
    // TODO: implement new methods in PredJTypeSystem.
    // TODO: override methods as needed from TypeSystem_c.

    protected TheoremProverInterface prover = new CVC3Interface_c(this);

	// NOTE: This method is only added to handle a bug in
	// Polyglot: it doesn't check that a method cannot be both
	// abstract and private.  I reported the bug, so this method
	// may later be unnecessary.
    public void checkMethodFlags(Flags f) throws SemanticException {
	super.checkMethodFlags(f.clear(PURE));
	if (f.isAbstract() && f.isPrivate()) {
	    throw new SemanticException(
		"Cannot declare method that is both abstract and private.");
	}
    }
    
    public Context createContext() {
	return new PredJContext_c(this);
    }

	// this is used to make the protected version of this method
	// in TypeSystem_c publicly available
    public boolean isAccessible(MemberInstance mi, ClassType ct) {
	return super.isAccessible(mi, ct);
    }

	// is this set of dispatcher mates overridden by a (concrete
	// or abstract) method of the same gf in the given class type?
    public boolean isOverridden(PredJDispatcherInstance di,
				ClassType ct) {
	PredJMethodInstance mi =
	    (PredJMethodInstance) di.dispatcherMates().get(0);
	if (!isAccessible(mi, ct) || mi.flags().isStatic()) {
	    return false;
	}
	List methods = ct.methods();
	Iterator iter = methods.iterator();
	while (iter.hasNext()) {
	    if (canOverride((MethodInstance) iter.next(), mi)) {
		return true;
	    }
	}
	return false;
    }
	
	

    public PredJFormalInstance formalInstance(Position pos, Flags flags,
					      Type type, String name,
					      int paramPos) {
        assert_(type);
	return new PredJFormalInstance_c(this, pos, flags, type, name,
					 paramPos);
    }


    public MethodInstance methodInstance(Position pos,
	                                 ReferenceType container, Flags flags,
					 Type returnType, String name,
					 List argTypes, List excTypes) {

        assert_(container);
        assert_(returnType);
        assert_(argTypes);
        assert_(excTypes);
	return new PredJMethodInstance_c(this, pos, container, flags,
					 returnType, name, argTypes, excTypes);
    }
    

    public ParsedClassType createClassType(LazyClassInitializer init,
					   Source fromSource) {
        return new PredJParsedClassType_c(this, init, fromSource);
    }    
    
    public PredJPredMethodInstance predMethodInstance(Position pos,
						      ReferenceType container,
						      Flags flags,
						      Type returnType,
						      String name,
						      List argTypes,
						      List excTypes,
						      Predicate predicate)
	{
	    assert_(container);
	    assert_(returnType);
	    assert_(argTypes);
	    assert_(excTypes);
	    return new PredJPredMethodInstance_c(this, pos, container, flags,
						 returnType, name, argTypes,
						 excTypes, predicate);
	}

    public PredJDispatcherInstance dispatcherInstance(
	List mates, boolean hasAbstractMethod) {
	return new PredJDispatcherInstance_c(mates, hasAbstractMethod,
					     this);
    }
	
    public Specializer createSpecializer(Predicate target,
					 Type specializer, boolean polarity) {
	return new Specializer_c(target, specializer, polarity);
    }   

    protected UnknownPredicate unknownPredicate = new UnknownPredicate_c(); 

    public UnknownPredicate unknownPredicate() {
	return unknownPredicate;
    }

    protected TruePredicate truePredicate = new TruePredicate_c(this); 

    public TruePredicate truePredicate() {
	return truePredicate;
    }

    public PredicateTarget createPredicateTarget(String name, Type type) {
	return new PredicateTarget_c(name, type);
    }

    public PredicateCall createPredicateCall(String name, List arguments,
                                             Type returnType,
                                             List argumentTypes) {
        return new PredicateCall_c(name, arguments, returnType, argumentTypes);
    }
    
    public PredicateLit createPredicateLit(Object value, Type type) {
	return new PredicateLit_c(value, type);
    }

    public PredicateBinary createPredicateBinary(
	Predicate left, Binary.Operator op, Predicate right, Type type) {
	return new PredicateBinary_c(left, op, right, type);
    }

    public PredicateArrayAccess createPredicateArrayAccess(
	Predicate array, Predicate index, Type type) {
	return new PredicateArrayAccess_c(array, index, type);
    }
    
    public PredicateUnary createPredicateUnary(
	Unary.Operator op, Predicate base, Type type) {
	return new PredicateUnary_c(op, base, type);
    }

    public PredicateSpecial createPredicateSpecial(
	Special.Kind kind, Type type) {
	return new PredicateSpecial_c(kind, type);
    }
    
    public TheoremProverInterface prover() {
	return prover;
    }

    	// is the method's predicate satisfiable?
    public boolean satisfiablePredicate(PredJMethodInstance m) {
	if (!m.hasPredicate()) {
	    return true;
	}

	Predicate notp = createPredicateUnary(Unary.NOT, predicate(m),
					      Boolean());
	return !prover().isTautology(notp);
    }

	// does one dispatcher mate override another?
    public boolean overrides(PredJMethodInstance mi,
			     PredJMethodInstance mj) {
	if (!mj.hasPredicate()) {
	    return true;
	}
	return prover().implies(predicate(mi), predicate(mj));
    }

	// get a list of the concrete methods that are part of the given
	// method's gf and are accessible from the given class.  these
	// are the set of methods that define this class's view of the gf,
	// which is used for both typechecking and codegen of the class.
    public List accessibleConcreteMethods(ClassType ct,
					  MethodInstance mi) {
	List allMIs = mi.implemented();
	    // remove methods that are not accessible; they cannot help
	    // ct be complete for this gf
	Iterator i = allMIs.iterator();
	while (i.hasNext()) {
	    if (! isAccessible((MethodInstance) i.next(), ct)) {
		i.remove();
	    }
	}

	    // remove all static methods that are not declared in the
	    // given class.  they are not really part of the same gf.
	    // incidentally, for some reason, Java does consider
	    // static methods in different classes to have an
	    // overriding relationship; e.g.  they must have the same
	    // return type.  not sure why this is necessary.  but in any
	    // case, static methods *are* considered to be overridden
	    // by things like MethodInstance::overrides.
	i = allMIs.iterator();
	while (i.hasNext()) {
	    PredJMethodInstance pmi = (PredJMethodInstance) i.next();
	    if (pmi.flags().isStatic() && ! pmi.container().equals(ct)) {
		i.remove();
	    }
	}	
	
	    // remove all abstract methods, and keep track of the
	    // most-specific one declared in a class
	i = allMIs.iterator();
	PredJMethodInstance abstractMethod = null;
	while (i.hasNext()) {
	    PredJMethodInstance pmi = (PredJMethodInstance) i.next();
	    if (pmi.flags().isAbstract()) {
		if (!pmi.flags().isInterface()) {
		    if (abstractMethod == null ||
			pmi.container().descendsFrom(
			    abstractMethod.container())) {
			abstractMethod = pmi;
		    }
		}
		i.remove();
	    }			    			    
	}		
	
	    // now remove all methods that are overridden by the
	    // most-specific abstract method (if any)
	if (abstractMethod != null) {
	    i = allMIs.iterator();
	    while (i.hasNext()) {
		PredJMethodInstance pmi = (PredJMethodInstance) i.next();
		if (abstractMethod.container().descendsFrom(
		    pmi.container())) {
		    i.remove();
		}
	    }
	}
	
	return allMIs;
    }	
	
	// is the GF consisting of the given methods (a list of
	// MethodInstances) complete?
    public boolean isCompletelyImplemented(List mis) {
	    // because we are doing the checking modularly, we need the
	    // equivalent of a local default.  in logical terms, the
	    // disjunction of the predicates of the given methods must be
	    // a tautology.

	    // note that there is no need to generate a default
	    // predicate, or some such.  in fact, it would be unsafe
	    // (I think) to do so, that is to check that the
	    // disjunction is equivalent to this default instead of to
	    // TRUE.  in particular, we cannot use properties of the
	    // class hierarchy to determine completeness, because we
	    // only have partial knowledge of that hierarchy.

	    // create the disjunction of the methods' predicates
	Iterator i = mis.iterator();
	Predicate p = null;
	while (i.hasNext()) {
	    PredJMethodInstance mi = (PredJMethodInstance) i.next();
		// don't include this method if it's abstract
	    if (!mi.flags().isAbstract()) {
		Predicate miP = predicate(mi);
		if (p == null) {
		    p = miP;
		} else {
		    p = createPredicateBinary(p, Binary.COND_OR, miP,
					      Boolean());
		}
	    }
	}

	    // check whether the disjunction is a tautology
	return prover().isTautology(p);
    }


    	// are the two methods disjoint, i.e. never both applicable at once
    public boolean areDisjoint(PredJMethodInstance mi,
			       PredJMethodInstance mj) {
	    // if at least one does not have a predicate, then the
	    // methods are not disjoint (unless the other method's predicate
	    // is logically false, but we can safely ignore that)
	if (!(mi.hasPredicate() && mj.hasPredicate()))
	    return false;
	
	Predicate pi = predicate(mi);
	Predicate pj = predicate(mj);

	    // check whether the two predicates are mutually exclusive,
	    // i.e. that !(pi && pj) is true.
	    // we equivalently check that pi implies !pj
	Predicate notpj = createPredicateUnary(Unary.NOT, pj, Boolean());
	return prover().implies(pi, notpj);
    }
	

	// do the set of methods that override the two given ones
	// completely cover the intersection of the given ones?  we
	// assume that neither given mate overrides the other.  the
	// arg partialOrder maps dispatcher mate #i to a set of the
	// mate numbers (Integers) of methods that mate #i
	// (transitively) overrides.  the given array of mates are
	// indexed by their mate number.
    public boolean haveResolvingMethods(int mate1, int mate2,
					List[] partialOrder,
					PredJMethodInstance[] mates) {
	    // first find all mates that override both of the given ones,
	    // and create the disjunction of their predicates
	Integer integer1 = new Integer(mate1);
	Integer integer2 = new Integer(mate2);
	Predicate resolvers = null;
	for (int i = 0; i < partialOrder.length; i++) {
	    if (partialOrder[i].contains(integer1) &&
		partialOrder[i].contains(integer2)) {
		    Predicate p = predicate(mates[i]);
		    if (resolvers == null) {
			resolvers = p;
		    } else {
			resolvers =
			    createPredicateBinary(p, Binary.COND_OR,
						  resolvers, Boolean());
		    }
	    }
	}

	    // if there are no methods overriding both mates, then the
	    // ambiguity is not resolved
	if (resolvers == null)
	    return false;

	    // create the intersection of the given mates' predicates
	Predicate p1 = predicate(mates[mate1]);
	Predicate p2 = predicate(mates[mate2]);
	Predicate intersection = createPredicateBinary(p1, Binary.COND_AND,
						       p2, Boolean());
	    // check whether the potential resolvers do in fact cover the
	    // intersection of the given mates.
	return prover().implies(intersection, resolvers);
    }
				       
    
    
	// a method without a predicate implicitly has the predicate TRUE
	// this method should never be called with an abstract method instance
    public Predicate predicate(PredJMethodInstance mi) {
//	assert(!mi.flags().isAbstract());
	if (mi instanceof PredJPredMethodInstance)
	    return ((PredJPredMethodInstance) mi).predicate();
	return truePredicate();
    }

    protected Flags PURE;

    protected void initFlags() {
        super.initFlags();
        PURE = createNewFlag("pure",
                             Public().Private().Protected().Static().Final());
    }

    public Flags Pure() {
        return PURE;
    }

    public boolean canOverride(MethodInstance mi, MethodInstance mj) {
        // Cannot override pure method with a non-pure method.
        if (! mi.flags().contains(PURE) && mj.flags().contains(PURE))
            return false;
        return super.canOverride(mi, mj);
    }

}

