package polyglot.ext.predj.types;

import polyglot.types.*;
import java.util.*;

public class Specializer_c extends Predicate_c implements Specializer {

    protected Predicate target;
    protected boolean polarity;

    public Specializer_c(Predicate target, Type specializer,
			 boolean polarity) {
	super(specializer);
	this.target = target;
	this.polarity = polarity;
    }
    
    public Predicate target() {
	return target;
    }

    public Set specializers() {
	Set s;
	    // we should conservatively add the static type of the target
	    // as a specializer, so we can declare axioms about it.  as an
	    // optimization, if this specializer expression's polarity is
	    // positive, then the target will already be known to have the
	    // specialized type, so the declared static type is unnecessary.
	    // this optimization has a large effect, because it reduces the
	    // number of axioms necessary, including axioms involving
	    // disjunction (see assertTargetTypes in CVCLInterface_c.java),
	    // which are expensive.
	if (polarity) {
	    s = new HashSet();
	} else {	    
	    s = target.specializers();
	}
	s.add(type());
	return s;
    }
    
    public Set targets() {
	Set s = new HashSet();
	    // the target could be 'this', in which case we don't add it
	if (target instanceof PredicateTarget || target instanceof PredicateArrayAccess) {
	    s.addAll(target.targets());  // Added All & .targets()  -- JasonR
	}
	return s;
    }
    
    public Set names() {
	Set s = new HashSet();
	    // the target could be 'this', in which case we don't add it
	if (target instanceof PredicateTarget || target instanceof PredicateArrayAccess) {
	    s.addAll(target.names());  // Added All & .names()  -- JasonR
	}
	return s;
    }
}
