package polyglot.ext.predj.util;

import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.ext.predj.visit.PredJTranslator;

// make an expr that is identical to this one, but substitute any new
// var bindings with their definitions, substitute a canonical name
// for each formal, and optionally cast formals and locals to their
// specialized types.

public class canonicalize {

    public static Receiver canonicalize(Receiver r,
					PredJContext c, NodeFactory nf) {
	if (r instanceof Local) {
	    Local l = (Local) r;
	    String name = l.name();
	    LocalInstance li = l.localInstance();
	    if (li instanceof PredJFormalInstance) {
		// it's a formal, so substitute a canonical name, based on
		// the formal's position
		PredJFormalInstance fi = (PredJFormalInstance) li;
		int paramPos = fi.paramPosition();
		Expr newFormal = l.name(PredJTranslator.argName(paramPos,
							    l.type()));
		if (c.formalIsSpecialized(name)) {
			// cast the formal to its specialized type
		    Type specializedType = c.findVariableSilent(name).type();
		    newFormal =
			nf.Cast(null, nf.CanonicalTypeNode(null,
							   specializedType), 
				newFormal);
		}
		return newFormal;
	    } else {
		// it's a local var binding declared in the predicate
		// so substitute it for its definition
		Expr varBinding = c.getPredicateVarBinding(name);
		if (varBinding != null) {
		    varBinding = (Expr) canonicalize(varBinding, c, nf);
		    // cast it to the new binding's type, if
		    // the binding is also specialized
		    if (c.varBindingIsSpecialized(name)) {
			Type bindingType = c.findVariableSilent(name).type();
			varBinding = nf.Cast(null,
					     nf.CanonicalTypeNode(null,
								  bindingType),
					     varBinding);
		    }
		    return varBinding;
		} else {
		    throw new RuntimeException("The context knows nothing about"
					       + " local variable " + name
					       + ".");
		}
	    }
	} else if (r instanceof Field) {
	    Field f = (Field) r;
	    return f.target(canonicalize(f.target(), c, nf));
	} else if (r instanceof Call) {
	    Call call = (Call) r;
//TODO fix this
	return call.target(canonicalize(call.target(), c, nf));
	} else if (r instanceof Special) {
	    return r;
	} else if (r instanceof TypeNode) {
	    return r;
	} else {
	    return r;
	}
    }
}
