package polyglot.ext.predj.visit;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.ext.jl.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.*;
import polyglot.util.*;
import polyglot.frontend.Job;

import java.util.*;

/** Visitor that computes the partial order among dispatcher mates in
    a class. **/
public class DispatcherBuilder extends ContextVisitor {
    public DispatcherBuilder(Job job, TypeSystem ts, NodeFactory nf) {
	super(job, ts, nf);
    }

    protected Node leaveCall(Node n) throws SemanticException {
	if (n instanceof ClassBody) {

	    ClassBody node = (ClassBody) n;
	    PredJNodeFactory nf = (PredJNodeFactory) nodeFactory();
	
	    ArrayList l = new ArrayList(node.members());
	    // partition the current class's methods by dispatcher mates
	    while (! l.isEmpty()) {
		ClassMember cm = (ClassMember) l.get(0);
		if (cm instanceof PredJMethodDecl) {
		    PredJMethodDecl m = (PredJMethodDecl) cm;
		    List dispatcherMates = dispatcherMates(m, l);
		    l.removeAll(dispatcherMates);
			// if there is only one mate for this gf and the
			// method does not have a predicate, then there
			// is no need to create a dispatcher method
		    if (dispatcherMates.size() > 1 || m.hasPredicate()) {
			node = node.addMember(
			    nf.PredJDispatcher(
				dispatcherMates, (PredJTypeSystem) ts));
		    }
		} else {
		    l.remove(cm);
		}
	    }

	    return node;
	} else
	    return super.leaveCall(n);
    }

	// return all dispatcher mates of m in members
    protected List dispatcherMates(PredJMethodDecl m, List members) {
	List l = new ArrayList();
	Iterator i = members.iterator();
	while(i.hasNext()) {
	    ClassMember cm = (ClassMember) i.next();
	    if (cm instanceof PredJMethodDecl &&
		m.hasDispatcherMate((PredJMethodDecl) cm)) {
		l.add(cm);
	    }
	}
	return l;
    }

}
