package polyglot.ext.predj.visit;

import polyglot.visit.*;
import polyglot.ast.*;
import polyglot.ext.predj.ast.*;
import polyglot.types.*;
import polyglot.ext.predj.types.PredJTypeSystem;

import polyglot.frontend.Job;
import polyglot.ext.predj.util.toJavaExpr;
import java.util.*;



/** Visitor that translates PredJ AST nodes to Java AST nodes, for
    subsequent output. **/
public  class PredJEclipseVisitor extends ContextVisitor {


	//The Main Static Global Array Variable
	public static ArrayList<MethodFamily> methodFamilyArray;
	
	//Variable used for arguments
    public static String argStr = "arg$";

    //Constructor
	public PredJEclipseVisitor(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
		methodFamilyArray = new ArrayList<MethodFamily>();		//Initialize Static Global Array Variable
	}

	//Function used for selecting correct type for arguments
    protected static String type2String(Type t) {
		if (t.isReference()) {
		    return "Object";
		} else if (t.isNumeric()) {
		    return "Real";
		} else if (t.isBoolean()) {
		    return "Boolean";
		} else {
		    throw new RuntimeException("Unexpected type " + t +	" in a predicate.");
		}	
    }
    
    //Function used to generate expression for replacing arguments
	public static String argName(int formalNum, Type t) {
		return argStr + formalNum + "$" + type2String(t);
    }

    protected Node leaveCall(Node n) throws SemanticException {
		if (n instanceof PredJDispatcher)
		{
			//Initalize all variables for Node
			ArrayList<Integer> methodLinesList = new ArrayList<Integer>();		//Array for Lines
			ArrayList<String> methodPredicateList = new ArrayList<String>();	//Array for Predicates
			ArrayList<ArrayList<Integer>> precedenceList = new ArrayList<ArrayList<Integer>>();		//Array for Precedence
			ArrayList<String> argExpression = new ArrayList<String>();			//Array for Argument Expression to be Replaced
			
			PredJDispatcher pd = (PredJDispatcher)n;
			List mates = pd.dispatcherMates();
			PredJMethodDecl m = (PredJMethodDecl)mates.get(0);
			Iterator iter = m.formals().iterator();

			int count = 0;							//Keeps the count for the number of method parameters
			
			String[] formalArray = new String[m.formals().size()];		//Array of formals
			
			//Generate array of formal names and arguments
			while (iter.hasNext())
			{
				Formal f = (Formal)iter.next();				
				String newName = argName(count, f.type().type());	
				formalArray[count] = f.name();
				argExpression.add(newName);
				count++;        									
			}


			//Gets Formal Method Parameters all in one String
			String methodParameterString = "";
			for (int j = 0; j < count; j++)
			{
				if (j == 0)
				{
					methodParameterString += m.formals().get(j).toString();
				}
				else
				{
					methodParameterString += ", " + m.formals().get(j).toString();
				}
			}


			//Removes java.lang. from String types in parameters
			String javaLang = "java.lang.";
			String emptyString = "";
			methodParameterString = methodParameterString.replace((CharSequence)javaLang, (CharSequence)emptyString);

			//Check each method by size of partialOrder
			for(int j = 0; j < pd.partialOrder().length; j++){
				PredJMethodDecl currMate = (PredJMethodDecl) mates.get(j);
				
				boolean hasPred = currMate.hasPredicate();
				
				//If the method has a predicate
				if(hasPred){
					PredJPredMethodDecl predMate = (PredJPredMethodDecl)currMate;
					Expr whenExpr = (Expr)toJavaExpr.toJavaExpr(predMate.predNode().predicateExpr(), nf, (PredJTypeSystem) ts);
					String whenString = whenExpr.toString();
			
					//Replacing variable names with argExpression
					for (int k = 0; k < m.formals().size(); k++)
					{
						whenString = whenString.replace((CharSequence)argExpression.get(k).toString(), (CharSequence)formalArray[k]);
					}

					//Store Lines and Predicate Expressions
					methodLinesList.add(predMate.predNode().position().line());
					methodPredicateList.add(whenString);	
				}
				//else
				else
				{

					//So, add the line number of the line to the methodLinesList
					//and "" in place of its name.
					int defaultpos = currMate.position().line();
					methodLinesList.add(defaultpos);
					methodPredicateList.add("");
					
				}
			}

			
			//Getting and Saving Ancestors/Precedence ordering
			List[] ancestors = pd.partialOrder();
		
			//List for storing the list of ancestor/precedence for each methodFamily
			precedenceList = new ArrayList<ArrayList<Integer>>();

			//Getting Ancestor Precedence List for each Predicate
			for (int p = 0; p < ancestors.length; p++)
			{
				//ArrayList for keeping ancestor/precedence of a single predicate 
				ArrayList<Integer> ancestorList = new ArrayList<Integer>();
				Iterator ancestorIter = ancestors[p].iterator();

				while (ancestorIter.hasNext())
				{
					//Gather all ancestor/precedence for a predicate
					Integer ancestorId = (Integer)ancestorIter.next();
					ancestorList.add(ancestorId);
				}
				//Store the predicate's ancestorList into precedenceList
				precedenceList.add(ancestorList);

			}


			//Making a MethodFamily instance for this one specific class
			MethodFamily methodFamily = new MethodFamily(m.name() + "(" + methodParameterString + ")");


			//Instantiating variables to store into data structure
			String className = context().currentClass().toString();
			String methodName = m.name();



			//Putting together the data structure
			for (int j = 0; j < methodLinesList.size(); j++)
			{	
				//Create predicateInstance and add the predicate's ancestor/precedence List
				Predicate predicateInstance = new Predicate(className, methodName, methodParameterString, methodPredicateList.get(j), j, methodLinesList.get(j));
				ArrayList<Integer> predicateclassPredicateList = precedenceList.get(j);
				for (int k = 0; k < predicateclassPredicateList.size(); k++)
				{
					predicateInstance.addAncestor(predicateclassPredicateList.get(k));
				}
				
				//Set Default
				if(methodPredicateList.get(j) == ""){
					predicateInstance.setDefault();
				}

				//Add Predicate to methodFamily
				methodFamily.addPredicate(predicateInstance);
			}
			methodFamilyArray.add(methodFamily);

		}
	    return super.leaveCall(n);
    }
}

