package polyglot.ext.predj.visit;

import java.util.ArrayList;

public class Predicate 
{
	
	//Declare variables
	protected String className;					//Name of class
	protected String methodName;				//Name of method family
	protected String formalArguments;			//Method parameters
	protected String when;						//Predicate when expression
	protected int lineNum;						//Predicate line number
	protected int id;							//Predicate id
	protected ArrayList<Integer> ancestorIds;	//ArrayList of Ancestors
	protected boolean isDefault;				//If it is a default predicate
	
	
	//Constructor
	public Predicate(String className, String methodName, String formalArguments, String when, int id, int lineNum)
	{
		this.className = className;
		this.methodName = methodName;
		this.formalArguments = formalArguments;
		this.when = when;
		this.id = id;
		this.lineNum = lineNum;
		ancestorIds = new ArrayList<Integer>();
		isDefault = false;
	}
	
	//Returns the Predicate when Expression.  If it is the default, return <Default>
	public String toString()
	{
		if(isDefault){
			return "<Default>";
		}
		else{
			return when;
		}
	}
	
	//For debugging purposes, printing out member data
	public String toLongString()
	{
		return "ClassName: " + className + " Method: " + methodName + "(" + formalArguments + ") when: \"" + when + "\" ID: " + id + " LineNum: " + lineNum;
	}

	//Add ancestors and precedence
	public ArrayList<Integer> getAncestors(){ return ancestorIds;}
	public void addAncestor(int ancestorId){	ancestorIds.add(ancestorId);}

	
	//Get and Set for member variables
	public void setDefault(){isDefault = true;}
	public boolean getDefault(){return isDefault;}
	public int getId() { return id;	}
	public String getWhen() { return when; }
	public String getClassName() { return className;}
	public String getMethodName() { return methodName; }
	public String getFormalArguments() { return formalArguments; }
	public int getLineNumber() { return lineNum; }

}
